###############################################################################
# $Id: MS_formatter.pl.src,v 1.10 1997/03/04 18:49:11 tom Exp $
# $Source: /RMM/RCS/pvr/scsi/utils/MS_formatter.pl.src,v $
###############################################################################
# Library Media Changer Output Formatter
###############################################################################
BEGIN { unshift(@INC, qw;/opt/VML/admin/lib/SCSI;); }

# push alternate path to formatter 
unshift(@INC, $ENV{'SCSI_FORMATTER_PATH'});

require mc_common;

# send all errors, warnings and help info to standard output
# open(STDERR, ">&STDOUT");

###############################################################################
$usage =
"Usage: MS_formatter -v Vendor -m DeviceModel -r Revision -o {sense|select} 
-i InputFile [-p PageCode] [-F OutputFormat] [Name=Value]*

Flags:
------
-v	Library vendor ID
-m	Device model
-r	Product revision level
-o	SCSI command (Mode Sense|Mode Select)
-i	Mode sense data input file
-p	Page code.  The valid page codes are:
		0x1d	Element Address Assignment Page
		0x1e	Transport Geometry Parameters Page
		0x1f	Device Capabilities Page
		0x20	Vendor Unique Page
		0x00	Vendor Unique Page

	The default page code is 0x3f, All Pages.
-F	Output Format.  Possible values are:
		shell	- mode page data is display as a series of
			  name-value pairs, one per output line
		colon	- colon-delimited output
	The default format is descriptive (suitable for human reading).
-h	Lists selectable mode parameters for the specified device
Name=Value
	For the mode select command, mode parameters are selected by
	specifying the parameter name and value.
-x	Verbose flag.\n";
###############################################################################

$vendor="";
$device_model="";
$revision="";
$command="";
$input_file="";
$page_code="0x3f";
$oformat="";
$verbose=0;
$data_length=0;
$bytes_read=0;
$ifh = \*STDIN;
$ofh = \*STDOUT;

#############################################################
# parse program arguments
#############################################################
$i=0;
while ( @ARGV ) {
	$_ = shift;

	/-v(.*)/ && do { $vendor = ($1 ? $1 : shift); next; };
	/-m(.*)/ && do { $device_model = ($1 ? $1 : shift); next; };
	/-r(.*)/ && do { $revision = ($1 ? $1 : shift); next; };
	/-i(.*)/ && do { $input_file = ($1 ? $1 : shift); next; };
	/-o(.*)/ && do { $command = ($1 ? $1 : shift); next; };
	/-p(.*)/ && do { $page_code = ($1 ? $1 : shift); next; };
	/-F(.*)/ && do { $oformat = ($1 ? $1 : shift); next; };
	/-h/ && do { $help++; next; };
	/-x/ && do { $verbose++; next; };

	# unknown arg
	/-.*/ && dye(5, "unknown argument $_.\n$usage");

	# what is left are the attributes 
	@attrs[$i++]= split;
}

#############################################################
# trim spaces from end of args
#############################################################
$vendor =~ s/\s*$//;
$device_model =~ s/\s*$//;

print "vendor=$vendor, model=$device_model, revision=$revision\n"
	if $verbose > 1;
print "input_file=$input_file, page_code=$page_code, oformat=$oformat\n"
	if $verbose > 1;

#############################################################
# validate vendor and device model 
#############################################################
dye(5, "no vendor specified.\n$usage") if ( $vendor eq "" );
dye(5, "no device model specified.\n$usage") if ( $device_model eq "" );

#############################################################
# include any vendor-specific definitions and routines
#############################################################
$vendor_module = "$vendor" . ".pm";

foreach $prefix ( @INC ) {
        print "searching path " . "$prefix/$vendor_module" . "\n"
        if $verbose > 1;

        if ( -e "$prefix/$vendor_module" ) {
                print "found vendor module " . "$vendor" . ".pm" . "\n"
                if $verbose > 1;

                require $vendor . ".pm";
                $mc_vendor_pages = $vendor . "_pages";

                last;
        }
}

#############################################################
# process help request - display changeable mode parameters
# by vendor page
#############################################################
$help && do {
	my ($found) = 0;
	foreach $p ( keys %{$mc_vendor_pages} ) {
		if ( defined(${$mc_vendor_pages}{$p}{'changeable'} ) ) {
			print "\nPage [$p]\n";
                        print "-----------\n";
 			print join ("\n", 
				@{ ${$mc_vendor_pages}{$p}{'changeable'} }), 
				"\n";
			$found++;
		}
	}
	if ( !$found ) {
                print "No selectable parameters defined for ";
                print "$vendor $device_model\n";
        }
	exit(0);
};

#############################################################
# validate other args
#############################################################
dye(5, "no revision specified.\n$usage") if ( $revision eq "" );
dye(5, "no command specified.\n$usage") if ( $command eq "" );
dye(5, "no input file specified.\n$usage") if ( $input_file eq "" );
dye(5, "cannot read input file $input_file.\n$usage") if (! -r $input_file);

#############################################################
# determine validate page codes for this device
#############################################################
(@valid_pages) = keys %mc_common_pages;
push(@valid_pages, keys %{$mc_vendor_pages});

# add 'all pages' page code
push(@valid_pages, '0x3f'); 
print "valid pages= @valid_pages\n" if $verbose;

#############################################################
# validate specified page code for this device
#############################################################
dye(5, "invalid page code $page_code for $vendor $device_model")
	if !is_member($page_code, \@valid_pages);

###############################################################################
# NAME:		mc_get_page
# DESCRIPTION:	Returns a reference to the specified mode page structure
# INPUT ARGS:	
#		1. Page code
#		References global associative array %mode_pages
# RETURNS:	
#		Reference to page, or 0.
###############################################################################
sub mc_get_page {
	my ($page) = @_;
	my  ($ref);

	return ($mode_pages{$page});
}

###############################################################################
# NAME:		mc_new_page
# DESCRIPTION:	Subroutine to create a new mode page structure
#		All mode pages are stored in an associative array
#		called %mode_pages.  The key for each mode page is
#		the page code, and the value is another hash which
#		contains the page's attributes.
# INPUT ARGS:	
#		1. Page code
#		2. Page saveable status
# RETURNS:	
#		Reference to page, or 0.
###############################################################################
sub mc_new_page {
	my ($page, $saveable) = @_;
	my  ($ref);

	# check for existence
	dye(1, "Duplicate page code encountered.\n") if $mode_pages{$page};

	# create a private hash for this page
	$ref = {};

	# associate it with the page
	$mode_pages{$page} = $ref;

	# prime with page code and page saveable parameters
	$ref->{'code'} = $page;
	$ref->{'saveable'} = $saveable;

	return ($ref);
}

###############################################################################
# NAME:		mc_get_header
# DESCRIPTION:	Subroutine to parse the mode sense data header block.
#
# The mode sense header block is depicted below:
#
# Mode Sense Data Header
# +------------------------------------------+
# | Bit/Byte | 7 | 6 | 5 | 4 | 3 | 2 | 1 | 0 |
# +------------------------------------------+
# |     0    |   Mode Sense Data Length      |
# +------------------------------------------+
# |     1    |   Reserved                    |
# +------------------------------------------+
# |     2    |   Reserved                    |
# +------------------------------------------+
# |     3    |   Reserved                    |
# +------------------------------------------+ 
#
# INPUT ARGS:	
#		1. Reference to the input file handle	
# RETURNS:	
#		Mode sense data length, or 0.
###############################################################################
sub mc_get_header {
	my $ifh = shift;
	my ($mc_hdr_len) = 4;
	my ($mc_hdr) = "C" x $mc_hdr_len;
	my ($data_len) = 0;
	my ($rec, $r1, $r2, $r3, $nbytes);

	#
        # read the data header and extract the data length.
        #
        $nbytes = read ($ifh, $rec, $mc_hdr_len);
	if ( $nbytes != $mc_hdr_len ) {
		return 0;	# EOF or error
	}

        ($data_len, $r1, $r2, $r3) = unpack($mc_hdr, $rec);
        ($verbose > 1 ) && do {
		print "\nMode Sense Data Header Contents\n";
		print "-------------------------------\n";
                printf "Mode Sense Data Length=0x%lx\n", $data_len;
        };

	# subtract off the 3 reserved bytes in the header block.
	return ($data_len - 3);
}

###############################################################################
# NAME:		mc_get_data_preamble
# DESCRIPTION:	Subroutine to parse the mode sense data preamble
#
# The data page preamble is depicted below:
#
# Data Page Preamble
# +------------------------------------------+
# | Bit/Byte | 7 | 6 | 5 | 4 | 3 | 2 | 1 | 0 |
# +------------------------------------------+
# |     0    | PS|Res|	    Page Code	     |	
# +------------------------------------------+
# |		Page Length	     	     |	
# +------------------------------------------+
#
# INPUT ARGS:	
#		1. Reference to the input file handle	
# RETURNS:	
#		In a list context:
#			(@preamble)
#			Contents: Page-Saveable, Page Code, Page Length
#		Return 0 on encountering EOF or error
###############################################################################
sub mc_get_data_preamble {
	my $ifh = shift;
	my ($mc_preamble_len) = 2;
	my ($mc_preamble) = "C" x $mc_preamble_len;
	my ($rec, $data_byte0, $page_length, $saveable, $page, $nbytes);
	my (@preamble);

        # read the first two bytes of the mode data page 
        $nbytes = read ($ifh, $rec, $mc_preamble_len);
	if ( $nbytes != $mc_preamble_len ) {
		return 0;	# EOF or error
	}

	$bytes_read += $nbytes;

	# extract the "Parameters Saveable" and Page Code values.
        ($data_byte0, $page_length) = unpack($mc_preamble, $rec);

        $saveable = (0x80 & $data_byte0) == 0x80 ? 1 : 0;
        $page = 0x3f & $data_byte0;

        ( $verbose > 1 ) && do {
		print "\nData Preamble Contents\n";
		print "----------------------\n";
        	printf "Page Code=0x%02.2lx\n", $page;
        	printf "PS-Parameters Saveable=%lx\n", $saveable;
                printf "Page Length=0x%lx\n", $page_length;
        };

	return ($saveable, $page, $page_length);
}	

###############################################################################
# NAME:		mc_lookup
# DESCRIPTION:	Lookup a subroutine name of the specified function
#		type for the vendor, model, revision.  All functions
#		are stored in associative arrays as follows:
#
#		%mc_common_pages		subroutines common to all libraries
#		%mc_vendor_pages		vendor-specific subroutines 
#		%mc_model_subs		(vendor, model) specific subroutines 
#		%mc_rev_subs		(vendor, model, revision) specific subs
# INPUT ARGS:
#		1. Mode page (hash key)
#		2. Vendor ID
#		3. Device model
#		4. Product revision level
#		5. Function type
# RETURNS:
#		Name of subroutine or undefined
###############################################################################
sub mc_lookup {
	my ($page, $vendor, $model, $revision, $fn_type) = @_;
	my ($func);

	# common routines 
	$func = $mc_common_pages{$page}{$fn_type};
	if ( defined($func) ) {
		return $func;
	}

	# vendor-specific routines
	$func = ${$mc_vendor_pages}{$page}{$fn_type};

	return $func;
}

###############################################################################
# NAME:		mc_load_page
# DESCRIPTION:	Subroutine to create and read in the mode sense 
#		data for a specified page.
# INPUT ARGS:
#		1. Page code
#		2. Page saveable parameter
#		3. Reference to the input file handle	
#		4. Page length
#		5. Vendor ID
#		6. Device Model
#		7. Product revision level
# RETURNS:
#		Reference to loaded page, or 0
###############################################################################
sub mc_load_page {
	my ($page, $saveable, $ifh, $page_length,
		$vendor, $model, $revision) = @_;
	my ($ref, $nbytes, $page_data); 

	# check for existence
	dye(1, "attempt to redefine page $page.\n") if $mode_pages{$page};

	($ref) = &mc_new_page($page, $saveable);

	# read in page data
	$nbytes = read($ifh, $page_data, $page_length);
	if ( $nbytes != $page_length ) {
		return 0;	# EOF or error
	}

	$bytes_read += $nbytes;

	# lookup the parse function
	$parse_fn = mc_lookup($page, $vendor, $model, $revision, 'parse_fn');

	if ( !defined($parse_fn) ) {
		warn "ERROR: System Error: no parse handler for ",
			"vendor=$vendor, model=$model,\n",
			"revision=$revision, page $page\n" if $verbose;
		return 0;
	}

	print "invoking $parse_fn\n" if ( $verbose > 1 );
	&$parse_fn($ref, $page_data, $page_length);

	return $ref;
}

###############################################################################
# NAME:		print_page
# DESCRIPTION:	Debug subroutine used to print page attributes
#		
# INPUT ARGS:
#		1. Reference to initialized mode page structure
# RETURNS:
#		Undefined
###############################################################################
sub print_page {
	my ($ref) = shift;

	# validate the reference
	ref($ref) eq "HASH"
		|| dye(1, "expected page reference.\n");

	my (@keys) = sort keys %$ref;
	my ($attr);

	foreach $attr ( @keys) {
		print "\t$attr = $ref->($attr)\n";
	}
}

#############################################################
# Main program starts
#############################################################

#############################################################
# open the mode sense data input file
#############################################################
open(INPUT,"<$input_file") || 
	dye(1, "cannot open input file $input_file $!\n");

# assign a reference to the file handle
$ifh = \*INPUT;

#############################################################
# read in the mode sense header block and get data length
#############################################################
$data_length = mc_get_header($ifh);
dye(1, "null input data file.\n") if ( $data_length == 0 );

while ( $bytes_read != $data_length ) {

	#############################################################
	# read in a mode page preamble and determine it's page code
	#############################################################
	($saveable, $page, $page_length) = mc_get_data_preamble($ifh);

	# convert page code to hex string format
	$page = sprintf("0x%02.2lx", $page);

	print "code=$page_code, page=$page, length=$page_length\n" 
		if ($verbose > 1);

	#############################################################
	# parse the mode page
	#############################################################
	$page_ref = mc_load_page($page, $saveable, $ifh, $page_length,
				 $vendor, $device_model, $revision);
	# skip this page if no parse handler defined
	next if ( $page_ref == 0 );

	ref($page_ref) eq "HASH" || dye(1, "expected page reference.\n");

	print_page($page_ref) if $verbose > 2;

	##############################################################
	# MODE SENSE
	# print the page data, if selected, in the specified oformat 
	##############################################################
	if ( ($command eq "sense") && 
	     ($page_code eq "0x3f" || $page_code eq $page) ) {
	        $print_fn = mc_lookup($page, $vendor, $device_model, 
					$revision, 'print_fn');

	        # skip this page if no print handler defined
	        if ( !defined($print_fn) ) {
                    warn "ERROR: System Error: no print handler for ",
                        "vendor=$vendor, model=$model,\n",
                        "revision=$revision, page $page\n" if $verbose;
                    next;
	        }

                print "invoking $print_fn\n" if ( $verbose > 1 );
                &$print_fn($page_ref, $ofh, $oformat);
	}
	#############################################################
	# MODE SELECT
	# if page supported, invoke setattrs function
	#############################################################
	elsif ( ($command eq "select") && ($page_code eq $page) ) {
	        $setattrs_fn = mc_lookup($page, $vendor, $device_model, 
					$revision, 'setattrs_fn');

	        # skip this page if no setattrs handler defined
	        next if ( !defined($setattrs_fn) );

                print "invoking $setattrs_fn\n" if ( $verbose > 1 );
                &$setattrs_fn($page_ref, @attrs);
	}

	print "data length=$data_length, bytes read=$bytes_read\n" 
		if ( $verbose > 2 );
}

#############################################################
# If select, write all dirty pages 
#############################################################
if ( $command eq "select" ) {
	foreach $page ( keys %{$mc_vendor_pages} ) {

        	$write_fn = mc_lookup($page, $vendor, $device_model, 
					$revision, 'write_fn');

        	# skip this page if no write handler defined
        	next if ( !defined($write_fn) );

		# write only if page is dirty
		if ( ${$mc_vendor_pages}{$page}{'dirty'} ) {
        		print "invoking $write_fn\n" if ( $verbose > 1 );
        		&$write_fn(mc_get_page($page), @attrs);
		}
	}
}
