###############################################################################
# $Id: mc_common.pm.src,v 1.6 1997/03/04 18:51:20 tom Exp $
# $Source: /RMM/RCS/pvr/scsi/utils/mc_common.pm.src,v $
###############################################################################
# Library Media Changer Output Formatter 
# Common Data and Subroutines
# This file is designed to be require'd by a front end script.
###############################################################################

###############################################################################
# Defaults/Global Constants/Global Variables
###############################################################################

# Common subroutines
%mc_common_pages = (
	# Element Address Assignment Page
	'0x1d' => {
		changeable	=> undef,
		dirty		=> undef,
		parse_fn	=> "parse_page_0x1d",
		print_fn	=> "print_page_0x1d",
		write_fn	=> undef,
	},
	# Transport Geometry Descriptor Page
	'0x1e' => {
		changeable	=> undef,
		dirty		=> undef,
		parse_fn	=> "parse_page_0x1e",
		print_fn	=> "print_page_0x1e",
		write_fn	=> undef,
	},
	# Device Capabilities Page
	'0x1f' => {
		changeable	=> undef,
		dirty		=> undef,
		parse_fn	=> "parse_page_0x1f",
		print_fn	=> "print_page_0x1f",
		write_fn	=> undef,
	},
);

###############################################################################
# NAME:		is_member
# DESCRIPTION:  Check to see if an element is present in a list.
#
# INPUT:
#		1.  Element to check for (string)
#		2.  Reference to list
# OUTPUT ARGS:
#		None.
# RETURNS:
#		0 - if not present
#		1 - if present
###############################################################################
sub is_member {
	my ($which, $lref) = @_;

	foreach $el (@$lref) {	
		return 1 if $el eq $which;
	}
	return 0;
}

###############################################################################
# NAME:		dye
# DESCRIPTION:	Writes a formatted error message and exits
# INPUT ARGS:	
#		1. Exit code
#		2. Error string
#		
# RETURNS:	
#		program exit code
###############################################################################
sub dye { 
my ($exit_code, $err_string) = @_;
my ($nothing);

	SWITCH: {
		if ( $exit_code == 0 ) { $err_type = "No Error"; last SWITCH; }
		if ( $exit_code == 1 ) { $err_type = "System Error"; 
					 last SWITCH; }
		if ( $exit_code == 2 ) { $err_type = "Check Condition"; 
					 last SWITCH; }
		if ( $exit_code == 3 ) { $err_type = "Device Busy"; 
					 last SWITCH; }
		if ( $exit_code == 4 ) { $err_type = "Reservation Conflict"; 
					 last SWITCH; }
		if ( $exit_code == 5 ) { $err_type = "Invalid Argument"; 
					 last SWITCH; }
		$nothing = 1;
	}
	
	print "ERROR:$err_type:$err_string\n";

	exit($exit_code);
}

###############################################################################
# NAME:		parse_page_0x1d
# DESCRIPTION:	Subroutine to parse the mode sense data for
#		the Element Address Assignment Page, (code '0x1d'). 
# INPUT ARGS:
#		1. Reference to initialized mode page structure
#		2. Data buffer
#		3. Page length
# RETURNS:
#		Reference to the mode page or 0 on error. 
###############################################################################
sub parse_page_0x1d {
	my ($ref, $rec, $param_len) = @_;
	my ($trans_addr, $trans_no, $store_addr, $store_no);
	my ($impex_addr, $impex_no, $data_addr, $data_no, $rev);
	my ($params) = "n" x ($param_len/2);

	# validate the reference
	ref($ref) eq "HASH"
		|| die "ERROR: System Error: expected page reference.\n";

	# extract the mode parameters for this page
        ($trans_addr, $trans_no, $store_addr, $store_no,
         $impex_addr, $impex_no, $data_addr, $data_no, $resv)
                                = unpack($params, $rec);

	# store the attributes for this page
	$ref->{'trans_addr'} = $trans_addr;
	$ref->{'trans_no'} = $trans_no;
	$ref->{'store_addr'} = $store_addr;
	$ref->{'store_no'} = $store_no;
	$ref->{'impex_addr'} = $impex_addr;
	$ref->{'impex_no'} = $impex_no;
	$ref->{'data_addr'} = $data_addr;
	$ref->{'data_no'} = $data_no;

	return $ref;
}

###############################################################################
# NAME:		print_page_0x1d
# DESCRIPTION:	Subroutine to print the mode sense data for
#		the Element Address Assignment mode page (code '0x1d'). 
# INPUT ARGS:
#		1. Reference to mode page structure
#		2. Reference to the output file handle	
#		3. Output format specification
# RETURNS:
#		1 or die.
###############################################################################
sub print_page_0x1d {
	my ($ref, $ofh, $oformat) = @_;
	my ($trans_addr, $trans_no, $store_addr, $store_no);
	my ($impex_addr, $impex_no, $data_addr, $data_no, $rev);

	# validate the reference
	ref($ref) eq "HASH"
		|| die "ERROR: System Error: expected page reference.\n";

	# print the attributes for this page.
	if ( $oformat eq "" ) {
		printf $ofh "\nElement Address Assignment Page\n";
		printf $ofh "-------------------------------\n";

		print
                "Page Code				[$ref->{'code'}]\n";
                print
                "PS-Page Saveable			[$ref->{'saveable'}]\n";

		printf $ofh
		"First Transport Element Address		[%d]\n", 
			$ref->{'trans_addr'};
		printf $ofh
		"Number of Transport Element		[%d]\n", 
			$ref->{'trans_no'};
		printf $ofh
		"First Storage Element	 		[%d]\n",
			$ref->{'store_addr'};
		printf $ofh
		"Number of Storage Element		[%d]\n", 
			$ref->{'store_no'};
		printf $ofh
		"1st Import/Export Element Address	[%d]\n", 
			$ref->{'impex_addr'};
		printf $ofh
		"Number of Import/Export Element		[%d]\n", 
			$ref->{'impex_no'};
		printf $ofh
		"First Data Transport Element Address	[%d]\n", 
			$ref->{'data_addr'};
		printf $ofh
		"Number of Data Transport Element	[%d]\n", 
			$ref->{'data_no'};
	}
	elsif ( $oformat eq "colon" ) {
        	printf $ofh "0x1d:%d:%d:%d:%d:%d:%d:%d:%d:%d\n", 
			$ref->{'saveable'},
			$ref->{'trans_addr'},
			$ref->{'trans_no'},
			$ref->{'store_addr'},
			$ref->{'store_no'},
			$ref->{'impex_addr'},
			$ref->{'impex_no'},
			$ref->{'data_addr'},
			$ref->{'data_no'};	
	}
	elsif ( $oformat eq "attribute" ) {
		print $ofh "PageCode=$ref->{'code'}\n";
		print $ofh "ParamsSave=$ref->{'saveable'}\n";
		print $ofh "TransAddr=$ref->{'trans_addr'}\n";
		print $ofh "TransNo=$ref->{'trans_no'}\n";
		print $ofh "StoreAddr=$ref->{'store_addr'}\n";
		print $ofh "StoreNo=$ref->{'store_no'}\n";
		print $ofh "ImpexAddr=$ref->{'impex_addr'}\n";
		print $ofh "ImpexNo=$ref->{'impex_no'}\n";
		print $ofh "DataAddr=$ref->{'data_addr'}\n";
		print $ofh "DataNo=$ref->{'data_no'}\n";
	}

	return 1;
}

###############################################################################
# NAME:		parse_page_0x1e
# DESCRIPTION:	Subroutine to parse the mode sense data for
#		the Transport Geometry Parameters Page, (code '0x1e'). 
# INPUT ARGS:
#		1. Reference to initialized mode page structure
#		2. Data buffer
#		3. Page length
# RETURNS:
#		Reference to the mode page or 0 on error. 
###############################################################################
sub parse_page_0x1e {
	my ($ref, $rec, $param_len) = @_;
	my ($params) = "C" x $param_len;
	my ($data_byte2);
	my ($media_rotate, $member_no); 

	# validate the reference
	ref($ref) eq "HASH"
		|| die "ERROR: System Error: expected page reference.\n";

	# extract the mode parameters for this page
        ($media_rotate, $member_no) = unpack($params, $rec);

	# Mask off the rotate bit
        $media_rotate = (0x01 & $data_byte2) == 0x01 ? 1 : 0;

	# store the attributes for this page
	$ref->{'param_len'} = $param_len;
	$ref->{'media_rotate'} = $media_rotate;
	$ref->{'member_no'} = $member_no;

	return $ref;
}

###############################################################################
# NAME:		print_page_0x1e
# DESCRIPTION:	Subroutine to print the mode sense data for
#		the Transport Geometry Parameters mode page (code '0x1e'). 
# INPUT ARGS:
#		1. Reference to mode page structure
#		2. Reference to the output file handle	
#		3. Output format specification
# RETURNS:
#		1 or die.
###############################################################################
sub print_page_0x1e {
	my ($ref, $ofh, $oformat) = @_;
	my ($media_rotate, $member_no); 

	# validate the reference
	ref($ref) eq "HASH"
		|| die "ERROR: System Error: expected page reference.\n";

	# print the attributes for this page.
	if ( $oformat eq "" ) {
		printf  $ofh
		"\nTransport Geometry Parameters Page\n";
		printf  $ofh
		  "----------------------------------\n";

		print
                "Page Code				[$ref->{'code'}]\n";
                print
                "PS-Page Saveable			[$ref->{'saveable'}]\n";

		printf $ofh 
		"Parameter Length			[0x%02.2x]\n", 
			$ref->{'param_len'};
		printf $ofh 
		"Media Rotation 				[%s]\n", 
			( $ref->{'media_rotate'} ? "Yes":"No" );
        	printf $ofh 
		"Member Number In Transport Element Set	[0x%02.2x]\n", 
			$ref->{'member_no'};
	}
	elsif ( $oformat eq "colon" ) {
        	printf $ofh "0x1e:%d:0x%02.2x:%s:0x%02.2x\n", 
			$ref->{'saveable'}, 
			$ref->{'param_len'}, 
			( $ref->{'media_rotate'} ? "Yes":"No" );
			$ref->{'member_no'};
	}
	elsif ( $oformat eq "attribute" ) {
		print $ofh "PageCode=$ref->{'code'}\n";
		print $ofh "ParamsSave=$ref->{'saveable'}\n";
		print $ofh "ParamLen=$ref->{'param_len'}\n";
		print $ofh "MediaRotate=$ref->{'media_rotate'}\n";
		print $ofh "MemberNo=$ref->{'member_no'}\n";
	}

	return 1;
}


###############################################################################
# NAME:		parse_page_0x1f
# DESCRIPTION:	Subroutine to parse the mode sense data for
#		the Device Capabilities Page, (code '0x1f'). 
# INPUT ARGS:
#		1. Reference to initialized mode page structure
#		2. Data buffer
#		3. Page length
# RETURNS:
#		Reference to the mode page or 0 on error. 
###############################################################################
sub parse_page_0x1f {
	my ($ref, $rec, $param_len) = @_;
	my ($params) = "C" x $param_len;
	my ($data_byte2, $data_byte3, $data_byte4, $data_byte5);
	my ($data_byte6, $data_byte7, $data_byte8, $data_byte9);
	my ($data_byte10, $data_byte11, $data_byte12, $data_byte13);
	my ($data_byte14, $data_byte15, $data_byte16, $data_byte17);
	my ($data_byte18, $data_byte19);
	
	my ($storDT, $storIE, $storST, $StorMT);
	my ($mt_to_dt, $mt_to_ie, $mt_to_st, $mt_to_mt);
	my ($st_to_dt, $st_to_ie, $st_to_st, $st_to_mt);	
	my ($ie_to_dt, $ie_to_ie, $ie_to_st, $ie_to_mt);	
	my ($dt_to_dt, $dt_to_ie, $dt_to_st, $dt_to_mt);	

	my ($mt_x_dt, $mt_x_ie, $mt_x_st, $mt_x_mt);
	my ($st_x_dt, $st_x_ie, $st_x_st, $st_x_mt);	
	my ($ie_x_dt, $ie_x_ie, $ie_x_st, $ie_x_mt);	
	my ($dt_x_dt, $dt_x_ie, $dt_x_st, $dt_x_mt);	

	# validate the reference
	ref($ref) eq "HASH"
		|| die "ERROR: System Error: expected page reference.\n";

        # extract the mode parameters for this page
	($data_byte2, $data_byte3, $data_byte4, $data_byte5, 
	 $data_byte6, $data_byte7, $data_byte8, $data_byte9, 
	 $data_byte10, $data_byte11, $data_byte12, $data_byte13, 
	 $data_byte14, $data_byte15, $data_byte16, $data_byte17, 
	 $data_byte18, $data_byte19) = unpack($params, $rec);

	# Mask off capabilities bits
        $storDT = (0x08 & $data_byte2) == 0x08 ? 1 : 0;
        $storIE = (0x04 & $data_byte2) == 0x04 ? 1 : 0;
        $storST = (0x02 & $data_byte2) == 0x02 ? 1 : 0;
        $storMT = (0x01 & $data_byte2) == 0x01 ? 1 : 0;

        $mt_to_dt = (0x08 & $data_byte4) == 0x08 ? 1 : 0;
        $mt_to_ie = (0x04 & $data_byte4) == 0x04 ? 1 : 0;
        $mt_to_st = (0x02 & $data_byte4) == 0x02 ? 1 : 0;
        $mt_to_mt = (0x01 & $data_byte4) == 0x01 ? 1 : 0;

        $st_to_dt = (0x08 & $data_byte5) == 0x08 ? 1 : 0;
        $st_to_ie = (0x04 & $data_byte5) == 0x04 ? 1 : 0;
        $st_to_st = (0x02 & $data_byte5) == 0x02 ? 1 : 0;
        $st_to_mt = (0x01 & $data_byte5) == 0x01 ? 1 : 0;

        $ie_to_dt = (0x08 & $data_byte6) == 0x08 ? 1 : 0;
        $ie_to_ie = (0x04 & $data_byte6) == 0x04 ? 1 : 0;
        $ie_to_st = (0x02 & $data_byte6) == 0x02 ? 1 : 0;
        $ie_to_mt = (0x01 & $data_byte6) == 0x01 ? 1 : 0;

        $dt_to_dt = (0x08 & $data_byte7) == 0x08 ? 1 : 0;
        $dt_to_ie = (0x04 & $data_byte7) == 0x04 ? 1 : 0;
        $dt_to_st = (0x02 & $data_byte7) == 0x02 ? 1 : 0;
        $dt_to_mt = (0x01 & $data_byte7) == 0x01 ? 1 : 0;

        $dt_x_dt = (0x08 & $data_byte12) == 0x08 ? 1 : 0;
        $dt_x_ie = (0x04 & $data_byte12) == 0x04 ? 1 : 0;
        $dt_x_st = (0x02 & $data_byte12) == 0x02 ? 1 : 0;
        $dt_x_mt = (0x01 & $data_byte12) == 0x01 ? 1 : 0;

        $st_x_dt = (0x08 & $data_byte13) == 0x08 ? 1 : 0;
        $st_x_ie = (0x04 & $data_byte13) == 0x04 ? 1 : 0;
        $st_x_st = (0x02 & $data_byte13) == 0x02 ? 1 : 0;
        $st_x_mt = (0x01 & $data_byte13) == 0x01 ? 1 : 0;

        $ie_x_dt = (0x08 & $data_byte14) == 0x08 ? 1 : 0;
        $ie_x_ie = (0x04 & $data_byte14) == 0x04 ? 1 : 0;
        $ie_x_st = (0x02 & $data_byte14) == 0x02 ? 1 : 0;
        $ie_x_mt = (0x01 & $data_byte14) == 0x01 ? 1 : 0;

        $dt_x_dt = (0x08 & $data_byte15) == 0x08 ? 1 : 0;
        $dt_x_ie = (0x04 & $data_byte15) == 0x04 ? 1 : 0;
        $dt_x_st = (0x02 & $data_byte15) == 0x02 ? 1 : 0;
        $dt_x_mt = (0x01 & $data_byte15) == 0x01 ? 1 : 0;

        # store the attributes for this page
        $ref->{'storDT'} = $storDT;
        $ref->{'storIE'} = $storIE;
        $ref->{'storST'} = $storST;
        $ref->{'storMT'} = $storMT;

        $ref->{'mt_to_dt'} = $mt_to_dt;
        $ref->{'mt_to_ie'} = $mt_to_ie;
        $ref->{'mt_to_st'} = $mt_to_st;
        $ref->{'mt_to_mt'} = $mt_to_mt;

        $ref->{'st_to_dt'} = $st_to_dt;
        $ref->{'st_to_ie'} = $st_to_ie;
        $ref->{'st_to_st'} = $st_to_st;
        $ref->{'st_to_mt'} = $st_to_mt;

        $ref->{'ie_to_dt'} = $ie_to_dt;
        $ref->{'ie_to_ie'} = $ie_to_ie;
        $ref->{'ie_to_st'} = $ie_to_st;
        $ref->{'ie_to_mt'} = $ie_to_mt;

        $ref->{'dt_to_dt'} = $dt_to_dt;
        $ref->{'dt_to_ie'} = $dt_to_ie;
        $ref->{'dt_to_st'} = $dt_to_st;
        $ref->{'dt_to_mt'} = $dt_to_mt;

        $ref->{'mt_x_dt'} = $mt_x_dt;
        $ref->{'mt_x_ie'} = $mt_x_ie;
        $ref->{'mt_x_st'} = $mt_x_st;
        $ref->{'mt_x_mt'} = $mt_x_mt;

        $ref->{'st_x_dt'} = $st_x_dt;
        $ref->{'st_x_ie'} = $st_x_ie;
        $ref->{'st_x_st'} = $st_x_st;
        $ref->{'st_x_mt'} = $st_x_mt;

        $ref->{'ie_x_dt'} = $ie_x_dt;
        $ref->{'ie_x_ie'} = $ie_x_ie;
        $ref->{'ie_x_st'} = $ie_x_st;
        $ref->{'ie_x_mt'} = $ie_x_mt;

        $ref->{'dt_x_dt'} = $dt_x_dt;
        $ref->{'dt_x_ie'} = $dt_x_ie;
        $ref->{'dt_x_st'} = $dt_x_st;
        $ref->{'dt_x_mt'} = $dt_x_mt;

	return $ref;
}


###############################################################################
# NAME:		print_page_0x1f
# DESCRIPTION:	Subroutine to print the mode sense data for
#		the Device Capabilities mode page (code '0x1f'). 
# INPUT ARGS:
#		1. Reference to mode page structure
#		2. Reference to the output file handle	
#		3. Output format specification
# RETURNS:
#		1 or die.
###############################################################################
sub print_page_0x1f {
	my ($ref, $ofh, $oformat) = @_;
	my ($storDT, $storIE, $storST, $StorMT);
	my ($mt_to_dt, $mt_to_ie, $mt_to_st, $mt_to_mt);
	my ($st_to_dt, $st_to_ie, $st_to_st, $st_to_mt);	
	my ($ie_to_dt, $ie_to_ie, $ie_to_st, $ie_to_mt);	
	my ($dt_to_dt, $dt_to_ie, $dt_to_st, $dt_to_mt);	

	my ($mt_x_dt, $mt_x_ie, $mt_x_st, $mt_x_mt);
	my ($st_x_dt, $st_x_ie, $st_x_st, $st_x_mt);	
	my ($ie_x_dt, $ie_x_ie, $ie_x_st, $ie_x_mt);	
	my ($dt_x_dt, $dt_x_ie, $dt_x_st, $dt_x_mt);	

	# validate the reference
	ref($ref) eq "HASH"
		|| die "ERROR: System Error: expected page reference.\n";

	# print the attributes for this page.
	if ( $oformat eq "" ) {
		print $ofh "\nDevice Capabilities Page\n";
		printf  $ofh "------------------------\n";

		print
                "Page Code			[$ref->{'code'}]\n";
                print
                "PS-Page Saveable		[$ref->{'saveable'}]\n";

		printf $ofh "Transport Holds Medium		[%s]\n", 
			($ref->{'storMT'}?"Yes":"No");
                printf $ofh "Storage Holds Medium		[%s]\n",
                        ($ref->{'storST'}?"Yes":"No");
                printf $ofh "Port Holds Medium		[%s]\n",
                        ($ref->{'storIE'}?"Yes":"No");
                printf $ofh "Data Transfer Holds Medium	[%s]\n",
                        ($ref->{'storDT'}?"Yes":"No");

                printf $ofh "Transport -> Transport		[%s]\n",
                        ($ref->{'mt_to_mt'}?"Yes":"No");
                printf $ofh "Transport -> Storage		[%s]\n",
                        ($ref->{'mt_to_st'}?"Yes":"No");
                printf $ofh "Transport -> Port		[%s]\n",
                        ($ref->{'mt_to_ie'}?"Yes":"No");
                printf $ofh "Transport -> Data Transfer	[%s]\n",
                        ($ref->{'mt_to_dt'}?"Yes":"No");

		printf $ofh "Storage -> Transport		[%s]\n",
                        ($ref->{'st_to_mt'}?"Yes":"No");
                printf $ofh "Storage -> Storage		[%s]\n",
                        ($ref->{'st_to_st'}?"Yes":"No");
                printf $ofh "Storage -> Port			[%s]\n",
                        ($ref->{'st_to_ie'}?"Yes":"No");
                printf $ofh "Storage -> Data Transfer	[%s]\n",
                        ($ref->{'st_to_dt'}?"Yes":"No");

                printf $ofh "Port -> Transport		[%s]\n",
                        ($ref->{'ie_to_mt'}?"Yes":"No");
                printf $ofh "Port -> Storage			[%s]\n",
                        ($ref->{'ie_to_st'}?"Yes":"No");
                printf $ofh "Port -> Port			[%s]\n",
                        ($ref->{'ie_to_ie'}?"Yes":"No");
                printf $ofh "Port -> Data Transfer		[%s]\n",
                        ($ref->{'ie_to_dt'}?"Yes":"No");

		printf $ofh "Data Transfer -> Transport	[%s]\n",
                        ($ref->{'dt_to_mt'}?"Yes":"No");
                printf $ofh "Data Transfer -> Storage	[%s]\n",
                        ($ref->{'dt_to_st'}?"Yes":"No");
                printf $ofh "Data Transfer -> Port		[%s]\n",
                        ($ref->{'dt_to_ie'}?"Yes":"No");
                printf $ofh "Data Transfer -> Data Transfer	[%s]\n",
                        ($ref->{'dt_to_dt'}?"Yes":"No");

                printf $ofh "Transport <-> Transport		[%s]\n",
                        ($ref->{'mt_x_mt'}?"Yes":"No");
                printf $ofh "Transport <-> Storage		[%s]\n",
                        ($ref->{'mt_x_st'}?"Yes":"No");
                printf $ofh "Transport <-> Port		[%s]\n",
                        ($ref->{'mt_x_ie'}?"Yes":"No");
                printf $ofh "Transport <-> Data Transfer     [%s]\n",
                        ($ref->{'mt_x_dt'}?"Yes":"No");

		printf $ofh "Storage <-> Transport		[%s]\n",
                        ($ref->{'st_x_mt'}?"Yes":"No");
                printf $ofh "Storage <-> Storage		[%s]\n",
                        ($ref->{'st_x_st'}?"Yes":"No");
                printf $ofh "Storage <-> Port		[%s]\n",
                        ($ref->{'st_x_ie'}?"Yes":"No");
                printf $ofh "Storage <-> Data Transfer       [%s]\n",
                        ($ref->{'st_x_dt'}?"Yes":"No");

                printf $ofh "Port <-> Transport		[%s]\n",
                        ($ref->{'ie_x_mt'}?"Yes":"No");
                printf $ofh "Port <-> Storage		[%s]\n",
                        ($ref->{'ie_x_st'}?"Yes":"No");
                printf $ofh "Port <-> Port			[%s]\n",
                        ($ref->{'ie_x_ie'}?"Yes":"No");
                printf $ofh "Port <-> Data Transfer		[%s]\n",
                        ($ref->{'ie_x_dt'}?"Yes":"No");

		printf $ofh "Data Transfer <-> Transport	[%s]\n",
                        ($ref->{'dt_x_mt'}?"Yes":"No");
                printf $ofh "Data Transfer <-> Storage       [%s]\n",
                        ($ref->{'dt_x_st'}?"Yes":"No");
                printf $ofh "Data Transfer <-> Port		[%s]\n",
                        ($ref->{'dt_x_ie'}?"Yes":"No");
                printf $ofh "Data Transfer <-> Data Transfer	[%s]\n",
                        ($ref->{'dt_x_dt'}?"Yes":"No");
	}
	elsif ( $oformat eq "colon" ) {
        	printf $ofh 
	      "0x1f:%d:%s:%s:%s:%s:%s:%s:%s:%s:%s:%s:%s:%s:%s:%s:%s:%s:%s:%s:",
			 $ref->{'saveable'},
			($ref->{'storMT'}?"Yes":"No"),
			($ref->{'storST'}?"Yes":"No"),
			($ref->{'storIE'}?"Yes":"No"),
			($ref->{'storDT'}?"Yes":"No"),
			($ref->{'mt_to_mt'}?"Yes":"No"),
			($ref->{'mt_to_st'}?"Yes":"No"),
			($ref->{'mt_to_ie'}?"Yes":"No"),
			($ref->{'mt_to_dt'}?"Yes":"No"),
			($ref->{'st_to_mt'}?"Yes":"No"),
			($ref->{'st_to_st'}?"Yes":"No"),
			($ref->{'st_to_ie'}?"Yes":"No"),
			($ref->{'st_to_dt'}?"Yes":"No"),
			($ref->{'ie_to_mt'}?"Yes":"No"),
			($ref->{'ie_to_st'}?"Yes":"No"),
			($ref->{'ie_to_ie'}?"Yes":"No"),
			($ref->{'ie_to_dt'}?"Yes":"No"),
			($ref->{'dt_to_mt'}?"Yes":"No"),
			($ref->{'dt_to_st'}?"Yes":"No");
        	printf $ofh 
		"%s:%s:%s:%s:%s:%s:%s:%s:%s:%s:%s:%s:%s:%s:%s:%s:%s:%s\n", 
			($ref->{'dt_to_ie'}?"Yes":"No"),
			($ref->{'dt_to_dt'}?"Yes":"No"),
                        ($ref->{'mt_x_mt'}?"Yes":"No"),
                        ($ref->{'mt_x_st'}?"Yes":"No"),
                        ($ref->{'mt_x_ie'}?"Yes":"No"),
                        ($ref->{'mt_x_dt'}?"Yes":"No"),
                        ($ref->{'st_x_mt'}?"Yes":"No"),
                        ($ref->{'st_x_st'}?"Yes":"No"),
                        ($ref->{'st_x_ie'}?"Yes":"No"),
                        ($ref->{'st_x_dt'}?"Yes":"No"),
                        ($ref->{'ie_x_mt'}?"Yes":"No"),
                        ($ref->{'ie_x_st'}?"Yes":"No"),
                        ($ref->{'ie_x_ie'}?"Yes":"No"),
                        ($ref->{'ie_x_dt'}?"Yes":"No"),
                        ($ref->{'dt_x_mt'}?"Yes":"No"),
                        ($ref->{'dt_x_st'}?"Yes":"No"),
                        ($ref->{'dt_x_ie'}?"Yes":"No"),
                        ($ref->{'dt_x_dt'}?"Yes":"No");
	}
	elsif ( $oformat eq "attribute" ) {
		print $ofh "PageCode=$ref->{'code'}\n";
		print $ofh "ParamsSave=$ref->{'saveable'}\n";

		printf $ofh "storMT=%s\n", ($ref->{'storMT'}?"Yes":"No");
                printf $ofh "storST=%s\n", ($ref->{'storST'}?"Yes":"No");
                printf $ofh "storIE=%s\n", ($ref->{'storIE'}?"Yes":"No");
                printf $ofh "storDT=%s\n", ($ref->{'storDT'}?"Yes":"No");

                printf $ofh "MT2MT=%s\n", ($ref->{'mt_to_mt'}?"Yes":"No");
                printf $ofh "MT2ST=%s\n", ($ref->{'mt_to_st'}?"Yes":"No");
                printf $ofh "MT2IE=%s\n", ($ref->{'mt_to_ie'}?"Yes":"No");
                printf $ofh "MT2DT=%s\n", ($ref->{'mt_to_dt'}?"Yes":"No");

		printf $ofh "ST2MT=%s\n", ($ref->{'st_to_mt'}?"Yes":"No");
                printf $ofh "ST2ST=%s\n", ($ref->{'st_to_st'}?"Yes":"No");
                printf $ofh "ST2IE=%s\n", ($ref->{'st_to_ie'}?"Yes":"No");
                printf $ofh "ST2DT=%s\n", ($ref->{'st_to_dt'}?"Yes":"No");

                printf $ofh "IE2MT=%s\n", ($ref->{'ie_to_mt'}?"Yes":"No");
                printf $ofh "IE2ST=%s\n", ($ref->{'ie_to_st'}?"Yes":"No");
                printf $ofh "IE2IE=%s\n", ($ref->{'ie_to_ie'}?"Yes":"No");
                printf $ofh "IE2DT=%s\n", ($ref->{'ie_to_dt'}?"Yes":"No");

		printf $ofh "DT2MT=%s\n", ($ref->{'dt_to_mt'}?"Yes":"No");
                printf $ofh "DT2ST=%s\n", ($ref->{'dt_to_st'}?"Yes":"No");
                printf $ofh "DT2IE=%s\n", ($ref->{'dt_to_ie'}?"Yes":"No");
                printf $ofh "DT2DT=%s\n", ($ref->{'dt_to_dt'}?"Yes":"No");

                printf $ofh "MTXMT=%s\n", ($ref->{'mt_x_mt'}?"Yes":"No");
                printf $ofh "MTXST=%s\n", ($ref->{'mt_x_st'}?"Yes":"No");
                printf $ofh "MTXIE=%s\n", ($ref->{'mt_x_ie'}?"Yes":"No");
                printf $ofh "MTXDT=%s\n", ($ref->{'mt_x_dt'}?"Yes":"No");

		printf $ofh "STXMT=%s\n", ($ref->{'st_x_mt'}?"Yes":"No");
                printf $ofh "STXST=%s\n", ($ref->{'st_x_st'}?"Yes":"No");
                printf $ofh "STXIE=%s\n", ($ref->{'st_x_ie'}?"Yes":"No");
                printf $ofh "STXDT=%s\n", ($ref->{'st_x_dt'}?"Yes":"No");

                printf $ofh "IEXMT=%s\n", ($ref->{'ie_x_mt'}?"Yes":"No");
                printf $ofh "IEXST=%s\n", ($ref->{'ie_x_st'}?"Yes":"No");
                printf $ofh "IEXIE=%s\n", ($ref->{'ie_x_ie'}?"Yes":"No");
                printf $ofh "IEXDT=%s\n", ($ref->{'ie_x_dt'}?"Yes":"No");

		printf $ofh "DTXMT=%s\n", ($ref->{'dt_x_mt'}?"Yes":"No");
                printf $ofh "DTXST=%s\n", ($ref->{'dt_x_st'}?"Yes":"No");
                printf $ofh "DTXIE=%s\n", ($ref->{'dt_x_ie'}?"Yes":"No");
                printf $ofh "DTXDT=%s\n", ($ref->{'dt_x_dt'}?"Yes":"No");

	}

	return 1;
}

